* QL WORLD DIY TOOLKIT - heap memory routines
* Version 0.7, Copyright 1989 Simon N Goodwin.
*
initialise lea.l    define,a1       A1 -> extension details
           move.w   $110,a2         Fetch BP.INIT vector
           jmp      (a2)            Add these extensions
*
* LINKUP "file name string"   - adds extension code in the heap
*
linkup     move.w   $116,a2         Fetch CA.GTSTR vector
           jsr      (a2)            Put string at 0(A1,A6.L)
           bne.s    bad_return      Return if unsuccessful
           subq.w   #1,d3           Only one parameter?
           bne.s    bad_param       Return unless just one
*
           move.l   a1,a0           A0 is name offset
           moveq    #-1,d1          Channel owner = this task
           moveq    #1,d3           OPEN_IN (shared access)
           trap     #4              Parameter is A6 relative
           moveq    #1,d0           IO.OPEN trap key
           trap     #2              Try to open the file
           tst.l    d0              Was OPEN OK?
           bne.s    bad_return      If not, exit & report error
*
           move.l   (a6),a1         A1 is BASIC buffer offset
           moveq    #64,d2          There's room for 64 bytes
           moveq    #-1,d3          Allow plenty of time
           trap     #4              Buffer is A6 relative
           moveq    #71,d0          FS.HEADR
           trap     #3              Read the file header
           tst.l    d0              Was it OK?
           bne.s    bad_close       Abort otherwise
*
           move.l   (a6),a1         Retrieve buffer offset
           move.l   0(a1,a6.l),d1   Get file length from header
           moveq    #-15,d0         Presume 'bad parameter'
           btst     #0,d1           Check length is even
           bne.s    bad_close       Close & complain otherwise
           move.l   d1,d4           Save length of code
           move.l   a0,a4           Save channel ID
           moveq    #0,d2           Owner is permanent task
           moveq    #24,d0          MT.ALCHP trap key
           trap     #1              Allocate memory
           tst.l    d0              Did it work?
           bne.s    bad_close       If not, close file & report
*
           move.l   a0,a1           Set load address
           move.l   d4,d2           Retrieve file length
           exg.l    a4,a0           Retrieve channel ID
           moveq    #-1,d3          Allow infinite time
           moveq    #72,d0          FS.LOAD trap key
           trap     #3              Load the entire file
           tst.l    d0              Did it load OK?
           bmi.s    bad_load        If not, tidy up & report
*
           moveq    #2,d0           IO.CLOSE
           trap     #2              Close the file
           jmp      (a4)            CALL the code
*
bad_load   move.l   d0,d7           Save cause of death
           move.l   a4,a0           Find the memory
           moveq    #25,d0          MT.RECHP trap key
           trap     #1              De-allocate RAM
           bra.s    close_out       Close file & report
*
bad_close  move.l   d0,d7           Save error code
close_out  moveq    #2,d0           IO.CLOSE trap key
           trap     #2              Close the file
           move.l   d7,d0           Restore error code
           rts           
*
bad_param  moveq    #-15,d0         Set BAD PARAMETER report
bad_return rts
*
* DISCARD buffer_address   - deallocate memory on the heap
*
discard    move.w   $118,a2         Fetch CA.GTLIN vector
           jsr      (a2)            Get long integers
           bne.s    bad_return      Return if fetch fails
           subq.w   #1,d3           Is there just 1 parameter?
           bne.s    bad_param       Reject otherwise
           move.l   0(a1,a6.l),d1   Get parameter from RI stack
           bmi.s    bad_param       Parameter must be >= 0
           btst     #0,d1           Parameter must be even
           bne.s    bad_param       Complain if it is odd
           move.l   d1,a0           Use parameter as a pointer
           cmpi.l   #'Bufa',-(a0)   Check header for watermark
           beq.s    seems_ok        Only continue if it matches
           moveq    #-7,d0          Otherwise report NOT FOUND
           rts
seems_ok   clr.l    (a0)+           Scrub watermark
           moveq    #25,d0          MT.RECHP trap key
           trap     #1              De-allocate RAM
           rts
*
* address/error code = RESERVE(bytes,owner)   - reserve heap
*
reserve    move.w   $118,a2         Fetch CA.GTLIN vector
           jsr      (a2)            Get long integers
           bne.s    bad_return      Give up if fetch fails
           subq.w   #2,d3           Two parameters?
           bne.s    bad_param       Give up unless two
           move.l   0(a1,a6.l),d1   Get number of bytes
           ble.s    bad_param       At least 1 byte needed?
           move.l   4(a1,a6.l),d2   Owner -1 = self, 0 = QDOS
           addq.l   #2,a1           Tweak stack by (8-6) = 2
           move.l   a1,$58(a6)      Set BV.RIP for result
           moveq    #24,d0          MT.ALCHP
           trap     #1              Allocate RAM
           tst.l    d0
           bmi.s    return_fp       Return error code
           move.l   #'Bufa',-4(a0)  Identify this block
           move.l   a0,d0           Return A0 via D0
*
* Make D0.L into a 6 byte decimal in the space on the RI stack
*
return_fp  move.w   d0,d4           D4 will be exponent
           move.l   d0,d5           D5 will be mantissa
           beq.s    normalised      Zero is a trivial case
           move.w   #2079,d4        First guess at exponent
           add.l    d0,d0           Already normalised?
           bvs.s    normalised
           subq.w   #1,d4           No, halve exponent weight
           move.l   d0,d5           Double mantissa to match
           moveq    #16,d1          Try a 16 bit shift
*
normalise  move.l   d5,d0           Take copy of mantissa
           asl.l    d1,d0           Shift mantissa d1 places
           bvs.s    too_far         Overflow; must shift less
           sub.w    d1,d4           Correct exponent for shift
           move.l   d0,d5           New mantissa is more normal
too_far    asr.w    #1,d1           Halve shift distance
           bne.s    normalise       Try shift of 8, 4, 2 and 1
normalised move.l   $58(a6),a1      Get safe A1 value
           move.l   d5,2(a1,a6.l)   Stack mantissa
           move.w   d4,0(a1,a6.l)   Stack exponent
           moveq    #2,d4           Floating point result
job_done   moveq    #0,d0
           rts
*
define     dc.w     2               Two procedures
           dc.w     linkup-*
           dc.b     6,'LINKUP'
           dc.w     discard-*
           dc.b     7,'DISCARD'
           dc.w     0,1             One function
           dc.w     reserve-*
           dc.b     7,'RESERVE'
           dc.w     0
           end

