* QL WORLD DIY TOOLKIT - MOUSE/POINTER DRIVER + KEY EMULATION
* Links SuperBASIC and Qdos key queues to a PC serial mouse
* Mouse Systems 3 button or Microsoft 2 button variants
* Version 2.8, Copyright  1993 Simon N Goodwin
*
* Extensions over version 1.n:
*
* 2.2 Optional co-ordinate wrap-around at screen borders
* 2.7 Optional Keyboard emulation for cursors and buttons
* 2.8 PTR_FN% function to read pointer characteristics
*
* Extensions planned for version 3.n:
*
* Support for analogue joysticks via I2C and Amiga ports
*
ser_port    equ      1              Configure to one or two
ser_pointer equ      148+ser_port*4 152 for SER1, 156 for SER2
buttons     equ      2              Configure to two or three
qkeys       equ      1              Configure to zero or one
*
* Configuration error trapping for DEVPAC etc.
*
            ifne     buttons-2
            ifne     buttons-3
            fail                    Wrong number of buttons
            endc
            endc     
*
            ifne     ser_port-1
            ifne     ser_port-2
            fail                    Serial port 1 or 2 expected
            endc
            endc
*
** 2.0 ** Ensure QKEYS is valid (0 or 1); value is used later
*
            ifne     qkeys
            ifne     qkeys-1
            fail                    QKEYS 0 or 1 = key emulation
            endc
            endc
*
** 2.0 ** Key codes
*
left        equ      192            Cursor arrows
right       equ      200
up          equ      208
down        equ      216
button1     equ      32             Space
button2     equ      10             Enter
button3     equ      27             Escape
*
sv_keyq     equ      76             Key queue system variable
*
* Offsets of variables in interrupt linkage
*
link        equ      0              Interrupt link comes first
vector      equ      4              Offset of code address
marker      equ      8              Offset of "DIYn" signature
prefix      equ      12
*
* These variables are relative to PREFIX
*                          PTR_FN%
latest_x    equ      0        0     Current co-ordinates
latest_y    equ      2        1
limit_x     equ      4        2     Right margin limit
limit_y     equ      6        3     Top margin limit
step_x      equ      8        4     Counts per horizontal move
step_y      equ      10       5     Counts per vertical move
button_bits equ      12       6     Bits shadow each button
synchro     equ      14       7     Input byte # 1..3 or 1..5
initial     equ      15             Microsoft initial byte
serial_id   equ      16      8+9    Zero or serial channel ID
*
            ifne     qkeys
drift_x     equ      20      10     ** 2.0 ** Accumulated X drift
drift_y     equ      22      11     Accumulated Y drift for keys
key_flag    equ      24      12     Set to request key queueing
wrap_flag   equ      26      13     Set to allow wrap at edges
            endc
*
var_end     equ      20+qkeys*8     ** 2.0 ** 8 bytes key space
var_length  equ      prefix+var_end
*
start       lea      define,a1
            movea.w  $110.w,a2      BP.INIT vector
            jmp      (a2)
*
* PTR_ON asks is it already on? if so, return no error
*
point_on    bsr.s    find_pos2
            beq.s    it_worked
*
* Otherwise, allocate linkage area in common heap
*
            moveq    #var_length,d1
            moveq    #0,d2          Owned by SuperBASIC
            moveq    #24,d0         MT.ALCHP trap key
            trap     #1
            tst.l    d0             Did we get it?
            bne.s    oops
            lea.l    server,a2
            move.l   a2,vector(a0)  Provide code address
            move.l   d7,marker(a0)  Add signature
            move.l   #$1FF00FF,prefix+limit_x(a0)    * 511, 255
            move.l   #12*65536+24,prefix+step_x(a0)  *  12, 24
            move.l   a0,d6
*
* See if serial channel is open
*
            moveq    #0,d0          MT.INF
            trap     #1
            tst.l    ser_pointer(a0) Is SER input queue open?
            bne.s    link_it        Nearly finished
*
* If not, take note & open it; assume correct baud rate
*
            lea.l    stream,a0      Point at channel name
            moveq    #0,d1          Permanent ownership
            moveq    #1,d3          OPEN_IN
            moveq    #1,d0          IO.OPEN trap key
            trap     #2
            tst.l    d0
            bne.s    no_input
            movea.l  d6,a1
            move.l   a0,prefix+serial_id(a1)
link_it     movea.l  d6,a0          Restore A0, link address
            moveq    #28,d0         Set up MT.LPOLL trap key
trap1_out   trap     #1
it_worked   moveq    #0,d0
            rts
*
no_input    move.l   d6,a0          Deallocate the heap space
            move.l   d0,d7
            moveq    #25,d0         MT.RECHP trap key
            trap     #1
            move.l   d7,d0          Return OPEN error code
            rts
*            
point_off   bsr.s    find_pos2      Is our interrupt linked?
            bmi.s    no_error       If not, return at once
*
* Close serial port if we opened it
*
            move.l   serial_id(a4),d0
            beq.s    not_opened
            movea.l  d0,a0
            moveq    #2,d0          IO.CLOSE trap key
            trap     #2
*
* Remove interrupt server and deallocate linkage memory
*
not_opened  lea.l    -prefix(a4),a0
            moveq    #29,d0         MT.RPOLL trap key
            trap     #1
            moveq    #25,d0         MT.RECHP trap key
            bra.s    trap1_out
*
find_pos2   bsr.s    find_pos       Extra call allows trapping
oops        rts
*
** 2.8 ** This became a subroutine for BUTTON% and PTR_FN%
*
get_an_int  move.l   (a7)+,a4       
            movea.w  $112.w,a2      Vector gets words
            jsr      (a2)
            bne.s    oops
            subq.w   #1,d3          Ensure just one parameter
            bne.s    bad_param      Reject otherwise
            move.w   0(a1,a6.l),d3
            jmp      (a4)
*
* BUTTON%(0) reads all buttons and returns a code 0 to 7 (+)
* BUTTON%(1..3) gives 0 unless the specified button is pressed
*
* Put the state of all the buttons into D0; bit 0 = Button 1
*
button      bsr.s    get_an_int
read_all    bsr.s    find_pos
            move.w   button_bits(a4),d4 Check up to 16 buttons
            beq.s    ret_integer    None so parameter irrelevant
            tst.w    d3
            beq.s    ret_integer    Parameter 0, no filtering
            cmp.w    #buttons,d3
            bhi.s    bad_param
            subq.w   #1,d3          Number bits from 0
            btst     d3,d4
            bne.s    ret_integer    If choice set, return all
main_line   moveq    #0,d4
            bra.s    ret_integer    Stack space is allocated
*
* SYNCH needs just four lines and ten bytes of code
*
synch       bsr.s    find_pos
            moveq    #0,d4
            move.b   synchro(a4),d4
            bra.s    chk_integer
*
* XPOINT% and YPOINT% differ only in the offset they use
*
xpoint      moveq    #latest_x,d5
            bra.s    get_word
ypoint      moveq    #latest_y,d5
get_word    bsr.s    find_pos
            move.w   0(a4,d5.l),d4  Read co-ordinate variable
*
* Return the integer in D4.W to SuperBASIC via the RI stack
*
chk_integer moveq    #2,d1          Number of bytes needed
            movea.w  $11A.w,a2      Find the BV.CHRIX vector
            jsr      (a2)           Allocate RI space
            subq.l   #2,$58(a6)     Update BV.RIP
ret_integer move.l   $58(a6),a1     Get BV.RIP
            move.w   d4,0(a1,a6.l)  Stack the result
            moveq    #3,d4          Type = 16 bit Integer
no_error    moveq    #0,d0          No run-time error
            rts
*
* FIND_POS - point A4 at the parameters in the interrupt list
*
* LONG link address
* LONG code address
* LONG "DIYn" marker
* WORD latest_x          <- A4
* WORD latest_y 
* WORD limit_x 
* WORD limit_y 
* WORD cursor x_step 
* WORD cursor y_step 
* LONG 0 or ID if SER was opened
*
* Word parameter values are 16 bit signed, valid from 1..32767
* Result in A4, "DIYn" marker in D7; job ID in D1; Corrupts D0
* Returns NOT FOUND to prior caller if DIY server is not linked
* No need for supervisor mode if Qdos performs links?
* 
find_pos    moveq    #0,d0          MT.INF
            trap     #1
            move.l   #'DIY1'+qkeys,d7
            lea.l    60(a0),a4      Locate polled list
find_loop   move.l   (a4),d0
            beq.s    not_found      Pass error to prior caller
            movea.l  d0,a4
            cmp.l    marker(a4),d7  Check signature
            bne.s    find_loop
found_it    lea.l    prefix(a4),a4  Point at the data
            rts
*
not_found   addq.l   #4,a7          Discard one return address
            moveq    #-7,d0
            rts
bad_param   moveq    #-15,d0        BAD PARAMETER error
bad_exit    rts
*
** 2.8 ** PTR_FN% added to read all PTR variables
*
get_ptr_var bsr      get_an_int
            bmi.s    bad_param      Ban negative offsets
            cmp.w    #var_end/2,d3
            bcc.s    bad_param      Too much
            bsr.s    find_pos
            add.w    d3,d3          Count words
            move.w   0(a4,d3.w),d4  Fetch a variable
            bra.s    ret_integer
*
* PTR_MAX, PTR_INC, PTR_POS etc. share most of their code
* They differ only in the address for the stored value in A4
*
** 2.0 **
            ifne     qkeys
point_key   bsr.s    find_pos
            lea.l    key_flag(a4),a4     Skip to flags
            bra.s    setwords
            endc
*
point_inc   bsr.s    find_pos
            addq.l   #step_x,a4     Skip first four co-ordinates
            bra.s    setwords
*
point_box   lea.l    16(a3),a3      Ignore first two parameters
            cmpa.l   a3,a5
            bhi.s    bad_param      Error if A3>A5
*
point_max   bsr.s    find_pos
            addq.l   #limit_x,a4    Skip first two co-ordinates
            bra.s    setwords
*
point_at    bsr.s    find_pos
setwords    movea.w  $112.w,a2      Vector gets words
            jsr      (a2)
            bne.s    bad_exit
            subq.w   #2,d3          Test for two parameters
            bne.s    bad_param      Reject otherwise
            move.w   0(a1,a6.l),d1  Check X
            bmi.s    bad_param
            swap     d1             Move X value to high word
            move.w   2(a1,a6.l),d1  Check Y
            bmi.s    bad_param
            move.l   d1,(a4)        Store checked value
exit        rts                     Return zero D0 error code
*
* This is the clever bit - the MS/PC mouse interrupt server
*
server      move.l   ser_pointer(a6),d0
            beq.s    exit
            movea.l  d0,a2          Find SER2 input queue
            movea.w  $E2.w,a1       IO.QOUT vector
            lea.l    8+prefix(a3),a4  Find variables
next_byte   jsr      (a1)
            bne.s    exit           Nothing to process now
*
* 33333333333333333333333333333333333333333333333333333333333
*
            ifeq     buttons-3
            ext.w    d1
            addq.b   #1,synchro(a4) Count bytes 1 to 5
            move.b   synchro(a4),d5
            subq.b   #1,d5          Use 0 to 4 internally
            beq.s    button_test
            btst     #0,d5          Is this a Y delta?
            bne.s    x_byte         No, it's an easy X
*
* Process the third and fifth bytes, the Y deltas
*
            ifeq     qkeys
y_byte      move.w   latest_y(a4),d2
            sub.w    d1,d2
            bmi.s    clear_yout     Limit minimum to zero
            move.w   limit_y(a4),d1
            cmp.w    d1,d2
            bls.s    in_yrange
            move.w   d1,d2          Set current Y to limit
            endc
** 2.0 **
            ifne     qkeys
y_byte      bsr      vertical
            move.w   latest_y(a4),d2
            sub.w    d1,d2
            bpl.s    enough_y       Check the upper limit
            tst.w    wrap_flag(a4)  Are we wrapping?
            beq.s    clear_yout
            add.w    limit_y(a4),d2
            bra.s    in_yrange
*
enough_y    move.w   limit_y(a4),d1
            cmp.w    d1,d2          Over the top?
            bls.s    in_yrange
            tst.w    wrap_flag(a4)  Are we wrapping?
            beq.s    clip_y
            sub.w    d1,d2          Wrap co-ordinate down
            bra.s    in_yrange
clip_y      move.w   d1,d2          Set current Y to limit
            endc
*
in_yrange   move.w   d2,latest_y(a4)
            bra.s    check_cycle
clear_yout  clr.w    latest_y(a4)
check_cycle subq.b   #4,d5          Is this the fifth byte?
            bne.s    next_byte
clear_sync  clr.b    synchro(a4)    Await next sequence start
            bra.s    next_byte
*
button_test moveq    #-8,d0         Mask out three button bits
            and.b    d1,d0
            cmp.b    #128,d0        Are the others as expected?
            bne.s    clear_sync     No, so keep waiting
            moveq    #7,d0          Mask in three button bits
            and.w    d1,d0          D0.W is just the buttons
            move.b   button_code(pc,d0),d0
** 2.0 **
            ifne     qkeys
            bsr      button_key     Key in if necessary
            endc
*
            move.w   d0,button_bits(a4)
            bra.s    next_byte      Button code converted
*
button_code dc.b     7,5,3,1,6,4,2,0
*
* Apply second and fourth bytes, the X deltas
*
            ifeq     qkeys
x_byte      move.w   latest_x(a4),d0
            add.w    d1,d0
            bmi.s    clear_xout     Limit minimum to zero
            move.w   limit_x(a4),d1
            cmp.w    d1,d0
            bls.s    in_xrange
            move.w   d1,d0          Set current X to limit
in_xrange   move.w   d0,latest_x(a4)
            bra.s    next_byte
clear_xout  clr.w    latest_x(a4)
            bra.s    next_byte
            endc
** 2.0 **
            ifne     qkeys
x_byte      bsr.s    horizontal
            move.w   latest_x(a4),d0
            add.w    d1,d0
            bpl.s    enough_x       Check the upper limit
            tst.w    wrap_flag(a4)  Are we wrapping?
            beq.s    clear_xout
            add.w    limit_x(a4),d0
            bra.s    in_xrange
*
enough_x    move.w   limit_x(a4),d1
            cmp.w    d1,d0          Over the top?
            bls.s    in_xrange
            tst.w    wrap_flag(a4)  Are we wrapping?
            beq.s    clip_x
            sub.w    d1,d0          Wrap co-ordinate to left
            bra.s    in_xrange
clip_x      move.w   d1,d0          Set current X to limit
in_xrange   move.w   d0,latest_x(a4)
next_near   bra      next_byte      Not in short range now
clear_xout  clr.w    latest_x(a4)
            bra.s    next_near
            endc
            endc
*
* 22222222222222222222222222222222222222222222222222222222222
*
            ifeq     buttons-2
            moveq    #0,d2          Clear high bytes
            move.b   d1,d2
            addq.b   #1,synchro(a4) Count bytes 1 to 3
            move.b   synchro(a4),d0
            moveq    #-64,d1        Mask for top two bits
            cmp.b    #2,d0          Switch three ways
            beq.s    byte_two
*
            ifeq     qkeys
            bpl.s    byte_three
            endc
*
            ifne     qkeys
            bpl      byte_three     Too far now for BPL.S
            endc
*
byte_one    move.b   d2,initial(a4) Save byte 1, used later
*
* Check & process first byte of message, button code 192+
*
            and.b    d1,d2          Check top bits only
            cmp.b    d1,d2          Are others as expected?
            bne.s    clear_sync     No, so keep waiting
button_test moveq    #48,d0         Mask only valid button bits
            and.b    initial(a4),d0
            lsr.w    #5,d0          Shift to bit 0 and carry
            bcc.s    only_one
            or.b     #2,d0          Flag button 2 in bit 1
only_one    
** 2.0 **
            ifne     qkeys
            bsr      button_key
            endc
*
            move.w   d0,button_bits(a4)
            bra.s    next_byte
*
resynch     move.b   #1,synchro(a4) Reset count to one
            move.b   d2,initial(a4) Treat new byte as first
            bra.s    button_test
*            
byte_two    and.b    d2,d1          Check top bits
            cmp.b    #128,d1
            beq.s    valid_x
            bcc.s    resynch
clear_sync  clr.b    synchro(a4)    Await start of sequence
            bra.s    next_byte
*
valid_x     and.b    #63,d2         D2 is six LS bits
            moveq    #3,d1
            and.b    initial(a4),d1 D1 is two MS bits 
            lsl.w    #6,d1
            or.w     d2,d1
            ext.w    d1             D1.W is signed X delta
*
            ifeq     qkeys
            move.w   latest_x(a4),d0
            add.w    d1,d0
            bmi.s    clear_xout     Don't move left of zero
            move.w   limit_x(a4),d1
            cmp.w    d1,d0
            bls.s    in_xrange
            move.w   d1,d0          Set current X to limit
in_xrange   move.w   d0,latest_x(a4)
            bra.s    next_byte
clear_xout  clr.w    d0             Zero
            bra.s    in_xrange
            endc
** 2.0 **
            ifne     qkeys
            bsr      horizontal
            move.w   latest_x(a4),d0
            add.w    d1,d0
            bpl.s    enough_x       Check the upper limit
            tst.w    wrap_flag(a4)  Are we wrapping?
            beq.s    clear_xout
            add.w    limit_x(a4),d0
in_xrange   move.w   d0,latest_x(a4)
near_xrange bra      next_byte
*
enough_x    move.w   limit_x(a4),d1
            cmp.w    d1,d0          Over the top?
            bls.s    in_xrange
            tst.w    wrap_flag(a4)  Are we wrapping?
            beq.s    clip_x
            sub.w    d1,d0          Wrap co-ordinate down
            bra.s    in_xrange
clip_x      move.w   d1,d0          Set current Y to limit
            bra.s    in_xrange
clear_xout  clr.w    d0             Zero
            bra.s    near_xrange    Loop via local label
            endc
*
* Process Y delta
*
byte_three  and.b    d2,d1          Check top bits
            cmp.b    #128,d1
            beq.s    valid_y
            bcc.s    resynch
            bra.s    clear_sync
*
valid_y     and.b    #63,d2         D2 is six LS bits
            moveq    #12,d1
            and.b    initial(a4),d1
            lsl.w    #4,d1          D1 is two MS bits 
            or.w     d2,d1
            ext.w    d1             D1.W is signed Y delta
*
            ifeq     qkeys
            move.w   latest_y(a4),d0
            add.w    d1,d0          D0 is new candidate Y
            bmi.s    clear_yout     Stop at zero
            move.w   limit_y(a4),d1
            cmp.w    d1,d0
            bls.s    in_yrange
            move.w   d1,d0          Stop at limit
in_yrange   move.w   d0,latest_y(a4)
            bra.s    clear_sync
clear_yout  clr.w    latest_y(a4)
            bra.s    clear_sync
            endc
** 2.0 **
            ifne     qkeys
            bsr      vertical
            move.w   latest_y(a4),d0
            add.w    d1,d0          D0 is new candidate Y
            bpl.s    enough_y       Check the upper limit
            tst.w    wrap_flag(a4)  Are we wrapping?
            beq.s    clear_yout
            add.w    limit_y(a4),d0
            bra.s    in_yrange
*
enough_y    move.w   limit_y(a4),d1
            cmp.w    d1,d0          Over the top?
            bls.s    in_yrange
            tst.w    wrap_flag(a4)  Are we wrapping?
            beq.s    clip_y
            sub.w    d1,d0          Wrap co-ordinate down
            bra.s    in_yrange
clip_y      move.w   d1,d0          Set current Y to limit
*
in_yrange   move.w   d0,latest_y(a4)
near_sync   bra      clear_sync
clear_yout  clr.w    latest_y(a4)
            bra.s    near_sync
            endc                    QKEYS
            endc                    2 BUTTONS
*
** 2.0 ** Routines to move cursor vertically or horizontally
* On entry D1 contains the new move delta and is preserved
* Uses A4, A6; corrupts A5, A2, D0, D3, D7; IO.QIN hits A3
*
            ifne     qkeys
horizontal  tst.w    key_flag(a4)
            beq.s    no_keys        Key entry disabled
            move.l   sv_keyq(a6),d0
            beq.s    no_keys
            move.w   d1,d7          Save delta
            move.l   a2,d6          Save input queue pointer
            movea.l  d0,a2          A2 -> Key queue
            move.w   drift_x(a4),d3
            movea.w  $E0.w,a5       IO.QIN vector
            add.w    d1,d3          D3 is total delta 
            bmi.s    go_left
            moveq    #right,d1
*
more        sub.w    step_x(a4),d3  Try a move
            bcs.s    add_xout
            jsr      (a5)           Cursor right
            bra.s    more
*
go_left     moveq    #left,d1
less        add.w    step_x(a4),d3  Try a move
            bpl.s    sub_xout
            jsr      (a5)           Cursor left
            bra.s    less
*
sub_xout    sub.w    step_x(a4),d3
            bra.s    set_xdrift     Store remainder
add_xout    add.w    step_x(a4),d3
set_xdrift  move.w   d3,drift_x(a4)
            move.l   d6,a2          Restore serial queue
            move.w   d7,d1          Restore delta
no_keys     rts
*
vertical    tst.w    key_flag(a4)
            beq.s    no_keys
            move.l   sv_keyq(a6),d0
            beq.s    no_keys        No queue to be stuffed
            move.w   d1,d7
*
            ifeq     buttons-3
            neg.w    d1             Reverse Y direction
            endc
*
            move.l   a2,d6          Save SER queue address
            move.l   d0,a2          A2 -> Key queue
            movea.w  $E0.w,a5       A5 = IO.QIN vector
            move.w   drift_y(a4),d3
            add.w    d1,d3          D3 is total delta 
            bmi.s    go_up
            moveq    #down,d1
descend     sub.w    step_y(a4),d3  Try a move
            bcs.s    add_yout       Don't go too far
            jsr      (a5)
            bra.s    descend
*
go_up       moveq    #up,d1
ascend      add.w    step_y(a4),d3  Try a move
            bpl.s    sub_yout
            jsr      (a5)
            bra.s    ascend
*
sub_yout    sub.w    step_y(a4),d3
            bra.s    set_ydrift
add_yout    add.w    step_y(a4),d3
set_ydrift  move.w   d3,drift_y(a4)
            move.l   d6,a2          Restore serial queue
            move.w   d7,d1          Restore new delta
            rts
*
* Routine to translate buttons to key-codes; new buttons in D0
*
button_key  tst.w    d0             Any action?
            beq.s    wits_end
            tst.w    key_flag(a4)   Key emulation on?
            beq.s    wits_end
            move.l   sv_keyq(a6),d1 Is there a key queue?
            beq.s    wits_end       No queue now so ignore
            move.w   d0,d7
            move.l   a2,d6          Save -> Serial queue
            move.l   d1,a2          A2 -> Key queue
            movea.w  $E0.w,a5       A5 = IO.QIN vector
            move.w   button_bits(a4),d1
            btst     #0,d0
            beq.s    not1
            btst     #0,d1          Already down?
            bne.s    not1
            moveq    #button1,d1    Send new keypress
            jsr      (a5)           Ignore 'no room' error
not1        btst     #1,d0          Button two?
            beq.s    not2
            btst     #1,d1          Already down?
            bne.s    not2
            moveq    #button2,d1    Queue it
            jsr      (a5)
*
not2        
            ifne     buttons-2
            btst     #2,d0          Third button?
            beq.s    restore
            btst     #2,d1
            bne.s    restore
            moveq    #button3,d1    Try to queue it
            jsr      (a5)
            endc
*
restore     move.w   d7,d0          Restore new buttons
            movea.l  d6,a2          Restore input queue
wits_end    rts
            endc
*
stream      dc.w     6
            dc.b     'SER','0'+ser_port,'IR'
*
define      dc.w     6+qkeys        Six or seven procedures
            dc.w     point_off-*
            dc.b     7,'PTR_OFF'
            dc.w     point_on-*
            dc.b     6,'PTR_ON'
            dc.w     point_box-*
            dc.b     10,'PTR_LIMITS' 0 , 0 , X% , Y%
            dc.w     point_at-*
            dc.b     7,'PTR_POS'    X% , Y%
            dc.w     point_max-*
            dc.b     7,'PTR_MAX'    X% , Y%
            dc.w     point_inc-*
            dc.b     7,'PTR_INC'    X% , Y%
** 2.0 **
            ifne     qkeys
            dc.w     point_key-*
            dc.b     7,'PTR_KEY'    KeyFlag%, WrapFlag%
            endc
*
            dc.w     0
            dc.w     5              Five functions
            dc.w     xpoint-*
            dc.b     6,'PTR_X%'
            dc.w     ypoint-*
            dc.b     6,'PTR_Y%'    
            dc.w     get_ptr_var-*
            dc.b     7,'PTR_FN%'    
            dc.w     button-*
            dc.b     7,'BUTTON%'    
            dc.w     synch-*
            dc.b     6,'SYNCH%'
            dc.w     0            
*
            end

