* LAURENCE REEVES' MEM VARIANT
*
* This version of MEM is modified by Laurence Reeves to suit
* Minerva and add extra features like EOF at the end of an
* allocated buffer, and larger buffer-sizes (e.g. MEM1_200K).
* It was downloaded from TF Services bulletin board. SNG 14/1/93
*
* N.B: THIS WAS ASSEMBLED WITH GST MACRO ASSEMBLER, NOT DEVPAC
* AND EXPECTS TO GENERATE A LINK FILE, NOT DIRECTLY USABLE CODE.
*
        section mem

* file name = MEM[<id>[_<size>][K][R|T|P]]
* <id>
*       omitted = absolute memory starting at 0.
*       -32768..32767 = identifier number.
* <size>
*       0..32767 = size of requested area.
* K
*       if a "K" is put after the size, the allocation is in kilobytes (x1024).
* R|T|P
*       R = set allocation status release.
*       T or omitted = allocation status unchanged. Set release if 1st chan.
*       P = set allocation status retain.

* When the first channel to an <id> is opened, memory of the requested <size>
* is allocated and cleared to zero. On subsequent opens, <size> is checked to
* be less than or equal to the original one supplied.

* When the last channel to an <id> is closed, if there has been a 'P' channel
* opened to the memory and if, since the last such, no subsequent channel has
* opted for 'R', the memory block for <id> is retained, ready for further
* channels to be opened to it. In all other instances, it is released.

* Requests to MEM<id> channels will report EOF when their file positions reach
* the allocated <size>. File positioning is supported, and postioning outside
* the bounds 0..<size> will finish up with no error but the position at the
* appropriate extreme.

mt.alchp equ    $18
mt.rechp equ    $19
mt.liod equ     $20

err.om  equ     -3
err.ef  equ     -10
err.bp  equ     -15

sd.extop equ    9

fs.posab equ    $42
fs.posre equ    $43

mm.alchp equ    $c0
mm.rechp equ    $c2
io.serio equ    $ea
io.name equ     $122

ch_end  equ     $18

* Driver definition block

d_link  equ     $18     L
d_io    equ     $1c     L       dc.l io,open,close
d_ser   equ     $28     L       dc.l pend,fbyte,sbyte
d_rts   equ     $34     L       rts
* spare equ     $36     W*3
d_name  equ     $3c     W+B*4   3,'MEM0'
* spare equ     $42     W
d_next  equ     $44     L       link all memory blocks together by m_id
d_end   equ     $48

* Memory block header

m_use   equ     0       W       no. of channels using this (msb set if perm)
m_id    equ     2       W       id of this block
m_next  equ     4       L       next memory block m_id
m_size  equ     8       L       requested size (msb set for dummy absolute)
m_end   equ     12

* Channel definition block

c_off   equ     ch_end+0 L      offset to current position
c_link  equ     ch_end+4 L      link to memory block m_size
c_end   equ     ch_end+8

        moveq   #d_end-d_link,d1
        bsr.s   mtalchp
        bne.s   rts0
        move.l  a0,a1
        lea     reloc,a2
rellp
        move.l  a3,(a1)+
        move.l  a2,a3
        move.w  (a2)+,d1
        sub.w   d1,a3
        bmi.s   rellp
        move.w  d1,(a1)+
        addq.l  #6,a1
        addq.w  #3,(a1)+
        move.l  mem,(a1)
        moveq   #mt.liod,d0
        bra.s   trap1

reloc   dc.w    *-io
        dc.w    *-open
        dc.w    *-close
        dc.w    *-pend
        dc.w    *-fbyte
        dc.w    *-sbyte
rts0
        rts     N.B. positive, acting as end marker for the above -ve's.

open
        lea     d_next(a3),a4   save block linkage address
        subq.l  #4*2,a7         space for parameters
        move.l  a7,a3
        move.w  (a0),d7         remember length of name
        move.w  io.name,a2
        jsr     (a2)
        bra.s   popex
        bra.s   popex
        bra.s   mmalchp
        dc.w    3
mem     ; use "MEM0" from here to set up ddb
        dc.w    'MEM0',4
abs     ; pointer to negative, used to make absolute MEM easy
        dc.w    -1,0,' _',0,1,'K',3,'RTP'

mtalchp
        moveq   #0,d2           allocate to job 0
        moveq   #mt.alchp,d0
trap1
        trap    #1
        tst.l   d0
        rts

mmalchp
        moveq   #c_end,d1
        move.w  mm.alchp,a2     get channel area
        jsr     (a2)
        beq.s   gotchp
popex
        addq.l  #4*2,a7         discard parameters
        rts

gotchp
        lea     abs,a1
        move.l  a1,c_link(a0)   make it look like an absolute block first
        subq.l  #3,d7           was name JUST "MEM"?
        beq.s   popex           yes - then we're finished

        movem.w (sp)+,d4-d7     get parameters d4=id/d5=size/d6=kilo/d7=perm
        mulu    #10,d6
        lsl.l   d6,d5
        bmi.s   bp_drop         don't let silly sizes through
lkup
        move.l  (a4),d1
        beq.s   newid           not found, go set up a new block
        move.l  d1,a4
        cmp.w   (a4)+,d4        is this the right id?
        bne.s   lkup            no - keep looking

        lea     m_size-m_next(a4),a5
        cmp.l   (a5),d5         is the size adequate?
        bls.s   idok            yes - go finish off
bp_drop
        moveq   #err.bp,d0
mmrechp
        move.l  d0,d5
        move.w  mm.rechp,a2     release channel
        jsr     (a2)
        move.l  d5,d0
        rts

newid
        moveq   #m_end+1,d1
        add.l   d5,d1           include space for header and a byte to ease i/o
        move.l  a0,a5           save channel pointer
        bsr.s   mtalchp
        exg     a5,a0           restore channel pointer
        bne.s   mmrechp
        addq.l  #m_id-m_use,a5
        move.w  d4,(a5)         set id
        move.l  a5,(a4)         tack onto end of linked list
        addq.l  #m_size-m_id,a5
        move.l  d5,(a5)         set size
idok
        move.l  a5,c_link(a0)   set pointer to memory block
        subq.l  #m_size-m_use,a5
        addq.w  #1,(a5)         one more channel in use
        asl.b   #7,d7           check <none>|R|T|P
        bpl.s   rts5            not R|P
        rol     (a5)            X=0|1 for R|P
        roxr    (a5)            replace release/retain flag
rts5
        rts

close
        lea     d_next(a3),a4   save dddb linkage address
        move.l  c_link(a0),a5   get pointer to memory block m_size
        bsr.s   mmrechp
        tst.b   (a5)            is this an absolute channel?
        bmi.s   rts8            yes - done
        subq.l  #m_size-m_id,a5
        move.l  a5,a0
        subq.w  #1,-(a0)        is this the last channel, and no retain flag?
        bne.s   rts8            no - leave data block allocated
find
        move.l  a4,a3           remember link address
        move.l  (a4),a4
        cmp.l   a4,a5           have we found this block yet?
        addq.l  #m_next-m_id,a4
        bne.s   find            no - keep looking
        move.l  m_next-m_id(a5),(a3) unlink the memory block
        moveq   #mt.rechp,d0    release memory block
        trap    #1
rts8
        rts

io
        moveq   #-sd.extop,d4
        add.b   d0,d4
        beq.s   extop
        move.l  c_link(a0),a2   a2 = data block size pointer
        sub.b   #fs.posab-sd.extop,d4
        beq.s   posab
        subq.b  #fs.posre-fs.posab,d4
        beq.s   posre
        move.l  (a2)+,d3        d3 = size, a2 = data block data
        bpl.s   normal
        sub.l   a2,a2           for absolute blocks, make base address zero
normal
        pea     d_ser(a3)       a3 = dddb address
        move.w  io.serio,a3
        move.l  a3,-(sp)
        move.l  c_off(a0),a3    a3 = running offset
        rts

posre
        add.l   c_off(a0),d1    make d1 into the absolute position required
        bvc.s   posab
        moveq   #-1,d1
        lsr.l   #1,d1
posab
        tst.l   d1
        bpl.s   posp
        moveq   #0,d1           negative values become zero
posp
        cmp.l   (a2),d1         is is beyond allocated area?
        blo.s   posit
        move.l  (a2),d1         yes - chop it to max
posit
        move.l  d1,a3           return new position
setoff
        move.l  a3,c_off(a0)
setok
        moveq   #0,d0
        rts

sbyte
        move.b  d1,0(a2,a3.l)   we allocated a spare byte to allow this to work
fbyte
        moveq   #1,d0
pent
        move.b  0(a2,a3.l),d1
        cmp.l   d3,a3
        add.l   d0,a3
        bcs.s   setoff
        moveq   #err.ef,d0
        rts

pend
        moveq   #0,d0
        bra.s   pent

extop
        jmp     (a2)

        end

